﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestZoom
{
    /// <summary>
    /// Dialog zur Anzeige eines Pseudo-Live-Bildes
    /// </summary>
    public partial class DlgCameraLive
        : Form
    {
        #region Member

        /// <summary>
        /// Flag zur Pausierung des Live-Bildes
        /// </summary>
        private bool LiveIsRunning;

        /// <summary>
        /// die Kamera-Instanz
        /// </summary>
        private CameraConnection Connection;

        /// <summary>
        /// Das Kamerabild in Kamera-Größe
        /// </summary>
        private Bitmap ImageOriginal;

        /// <summary>
        /// Das mit aktueller Zoomstufe dargestellte Bild
        /// </summary>
        private Bitmap ImageView;

        /// <summary>
        /// Flag zur Anzeige zur Anpassung der Scroll-Properties
        /// </summary>
        private bool DoChangeScroll;

        /// <summary>
        /// Punkt zur Zentrierung der
        /// View-Mitte bei Zoom-Änderung
        /// </summary>
        private Point CenterPoint;

        #endregion Member

        #region Konstruktor

        /// <summary>
        /// Konstruktor der Klasse DlgCameraLive
        /// </summary>
        /// <param name="conn">die Kamera-Instanz</param>
        public DlgCameraLive(CameraConnection conn)
        {
            this.InitializeComponent();
            this.LiveIsRunning = false;
            this.Text = "Camera Live";
            this.Connection = conn;
            this.DistributeNewImage();
            double factorCam = (double)this.Connection.Width / (double)this.Connection.Height;
            double factorBox = (double)this.ClientSize.Width / (double)this.ClientSize.Height;
            this.ClientSize = new System.Drawing.Size(this.ClientSize.Width, (int)(this.ClientSize.Height * factorBox / factorCam));
        }

        #endregion Konstruktor

        #region Public Stuff

        /// <summary>
        /// Starten und Anhalten des Live-Bildes
        /// </summary>
        /// <param name="run"></param>
        public void RunLive(bool run)
        {
            this.pbImage.Pause = !run;
            this.pbImage.Invalidate();
            this.DistributeNewImage();
            this.timer1.Enabled = run;
        }

        #endregion Public Stuff

        #region Private Stuff

        /// <summary>
        /// Holt eines neues Bildes von der Kamera
        /// und verteilt dies an die Dialoge
        /// </summary>
        private void DistributeNewImage()
        {
            Bitmap bmp = this.Connection.Snap();
            if (this.ImageOriginal != null)
            {
                this.ImageOriginal.Dispose();
            }

            this.ImageOriginal = bmp;

            if (bmp == null)
            {
                this.pbImage.SetImage(null);
                return;
            }

            this.CreateZoomImage();
        }

        /// <summary>
        /// Das Kamerabild wird entsprechend der Zoomstufe
        /// in ImageView skaliert und dargestellt
        /// </summary>
        private void CreateZoomImage()
        {
            if (this.ImageOriginal == null)
            {
                return;
            }

            if (this.ImageView != null)
            {
                this.ImageView.Dispose();
            }

            // Übergroßes Bild
            if (this.nudZoom.Value > 1)
            {
                try
                {
                    // hier wird ggf. ein überdimensionales Bild erzeugt!!!
                    // hier kam eine OutOfMemory-Exception
                    Bitmap bmp = DlgCameraLive.CreateZoomImage(this.ImageOriginal, (int)this.nudZoom.Value);
                    this.ImageView = bmp;

                    if (bmp == null)
                    {
                        Console.Out.WriteLine("Out Of Memory, set Zoom Value to minimum");
                        this.nudZoom.Value = this.nudZoom.Minimum;
                        return;
                    }
                }
                catch
                {
                    // OutOfMemoryException, ApplicationException
                    //Console.Out.WriteLine(ex.Message);
                    Console.Out.WriteLine("OutOfMemoryException, set Zoom Value to minimum");
                    this.nudZoom.Value = this.nudZoom.Minimum;
                    return;
                }
            }
            else
            {
                // Zoom = 0 - Darstellung des Bildes im Control
                // Zoom = 1 - 1 Kamera-Pixel = 1 Bildpixel
                this.ImageView = (Bitmap)this.ImageOriginal.Clone();
            }

            this.pbImage.Zoom = (int)this.nudZoom.Value;
            this.pbImage.SetImage(this.ImageView);

            if (this.nudZoom.Value == 0)
            {
                // SizeMode muss hier zuerst gesetzt werden!
                this.pbImage.SizeMode = PictureBoxSizeMode.Zoom;
                this.pbImage.Update();

                this.pnImage.AutoScroll = false;
                this.pnImage.VerticalScroll.Value = 0;
                this.pnImage.HorizontalScroll.Value = 0;
                this.pnImage.AutoScrollPosition = Point.Empty;

                this.pbImage.Update();

                // Vollbild, keine Scrollbar
                this.pbImage.Size = new Size(
                    this.pnImage.Size.Width - this.pnImage.Margin.Left - this.pnImage.Margin.Right - 1,
                    this.pnImage.Size.Height - this.pnImage.Margin.Top - this.pnImage.Margin.Bottom - 1);
            }
            else
            {
                this.pnImage.AutoScroll = true;
                // Scrollbars aktivieren
                this.pbImage.Size = this.ImageView.Size;
                this.pbImage.SizeMode = PictureBoxSizeMode.AutoSize;
            }

            if (this.DoChangeScroll)
            {
                this.ChangeScrollProperties();
            }
        }

        /// <summary>
        /// hier wird ggf. ein überdimensionales Bild erzeugt!!!
        /// Bereitstellen des gezoomten Bildes.
        /// </summary>
        /// <param name="bmpSource">die Vorlage</param>
        /// <param name="zoom">die Zoomstufe</param>
        /// <returns>das gezoomte Bild</returns>
        private static Bitmap CreateZoomImage(Bitmap bmpSource, int zoom)
        {
            // Graphics.FromImage() funktioniert nicht bei indizierten Bitmaps,
            // deswegen muss das Zoombild zu Fuß erstellt werden
            System.Diagnostics.Debug.Assert(bmpSource.PixelFormat == PixelFormat.Format8bppIndexed);

            int ww = bmpSource.Width;
            int hh = bmpSource.Height;

            // Hier kommt ggf. eine OutOfMemoryException oder ApplicationException
            Bitmap bmpZoom = new Bitmap(zoom * ww, zoom * hh, PixelFormat.Format8bppIndexed);

            byte[] bySource = new byte[ww];
            byte[] byZoom = new byte[zoom * ww];
            unsafe
            {
                // die Quell-Bitmap locken
                BitmapData dataSource = bmpSource.LockBits(new Rectangle(0, 0, ww, hh), ImageLockMode.ReadOnly, PixelFormat.Format8bppIndexed);
                // die Ziel-Bitmap locken
                BitmapData dataZoom = bmpZoom.LockBits(new Rectangle(0, 0, zoom * ww, zoom * hh), ImageLockMode.WriteOnly, PixelFormat.Format8bppIndexed);

                // Pointer auf den Anfang der Quell-Pixel
                byte* ptrSource = (byte*)dataSource.Scan0.ToPointer();
                // Pointer auf den Anfang der Ziel-Pixel
                byte* ptrZoom = (byte*)dataZoom.Scan0.ToPointer();

                // Schleife über alle Zeilen der Quell-Bitmap
                int rowZoom = 0;
                for (int sourceRow = 0; sourceRow < hh; sourceRow++)
                {
                    // Pointer auf den Start der aktuellen Zeile
                    byte* ptr1 = ptrSource + (sourceRow * dataSource.Stride);
                    // eine Zeile rauskopieren
                    Marshal.Copy(new IntPtr(ptr1), bySource, 0, ww);

                    // Zeile entsprechend Zoom vervielfältigen
                    int i2 = 0;
                    for (int i = 0; i < ww; i++)
                    {
                        for (int j = 0; j < zoom; j++)
                        {
                            byZoom[i2] = bySource[i];
                            i2++;
                        }
                    }

                    for (int j = 0; j < zoom; j++)
                    {
                        // Pointer auf den Start der aktuellen Zeile
                        byte* ptr2 = ptrZoom + (rowZoom * dataZoom.Stride);
                        // eine Zeile zurückkopieren
                        Marshal.Copy(byZoom, 0, new IntPtr(ptr2), zoom * ww);
                        rowZoom++;
                    }
                }

                // wieder freigeben
                bmpSource.UnlockBits(dataSource);
                // wieder freigeben
                bmpZoom.UnlockBits(dataZoom);
            }

            return bmpZoom;
        }

        /// <summary>
        /// Setzen der Scroll-Properties so,
        /// dass die Mitte des Bildes erhalten bleibt
        /// </summary>
        private void ChangeScrollProperties()
        {
            if (this.pnImage.HorizontalScroll.Visible)
            {
                int x = (int)(this.CenterPoint.X * this.nudZoom.Value);
                x -= this.pnImage.HorizontalScroll.LargeChange / 2;
                this.pnImage.HorizontalScroll.SetValue(x);
                this.pnImage.Update();
                this.pnImage.HorizontalScroll.SetValue(x);
                this.pnImage.Update();
            }

            if (this.pnImage.VerticalScroll.Visible)
            {
                int y = (int)(this.CenterPoint.Y * this.nudZoom.Value);
                y -= this.pnImage.VerticalScroll.LargeChange / 2;
                this.pnImage.VerticalScroll.SetValue(y);
                this.pnImage.Update();
                this.pnImage.VerticalScroll.SetValue(y);
                this.pnImage.Update();
            }

            this.DoChangeScroll = false;
            this.nudZoom.Enabled = true;
        }

        #endregion Private Stuff

        #region EventHandler

        /// <summary>
        /// nächstes Pseudo-Live-Bild anstoßen
        /// </summary>
        private void timer1_Tick(object sender, EventArgs e)
        {
            if (this.LiveIsRunning || this.DoChangeScroll)
            {
                // gerade ein Bild in Arbeit
                return;
            }

            try
            {
                this.LiveIsRunning = true;
                this.DistributeNewImage();
                this.LiveIsRunning = false;
            }
            catch
            {
                // Kameraverbindung verloren
                this.timer1.Stop();
                if (this.ImageView != null)
                {
                    this.ImageView.Dispose();
                    this.ImageView = null;
                }

                Console.Out.WriteLine("Camera connection lost.\nLive view stopped.");
                this.Close();
            }
        }

        private void DlgLive_FormClosing(object sender, FormClosingEventArgs e)
        {
            this.timer1.Stop();
        }

        private void nudZoom_ValueChanged(object sender, EventArgs e)
        {
            this.nudZoom.Enabled = false;
            this.DoChangeScroll = true;

            // View-Mitte in Panel-Koordinaten
            Point pt1 = new Point(this.pnImage.Width / 2, this.pnImage.Height / 2);
            Point pt2 = this.pnImage.PointToScreen(pt1);
            // View-Mitte in Bitmap-Koordinaten
            Point pt3 = this.pbImage.PointToClient(pt2);

            if (this.nudZoom.Value == 0)
            {
                this.pbImage.Zoom = 0;
                pt3 = pt1;
            }

            // View-Mitte in Bitmap-Koordinaten
            this.CenterPoint = this.pbImage.ScaleToImage(pt3);

            this.pbImage.Zoom = (int)this.nudZoom.Value;

            if (!this.LiveIsRunning)
            {
                this.CreateZoomImage();
            }
        }

        #endregion EventHandler
    }
}
